@@ -5,7 +5,8 @@ from pprint import pprint
 from trcli.api.api_client import APIClient, APIClientResult
 from trcli.cli import Environment
 from trcli.api.api_response_verify import ApiResponseVerify
-from trcli.data_classes.dataclass_testrail import TestRailSuite
+from trcli.data_classes.dataclass_testrail import TestRailSuite, TestRailCase
+from trcli.data_classes.data_parsers import MatchersParser
 from trcli.data_providers.api_data_provider import ApiDataProvider
 from trcli.constants import (
     ProjectErrors,
@@ -37,7 +38,7 @@ class ApiRequestHandler:
         self.environment = environment
         self.client = api_client
         self.suffix = api_client.VERSION
-        self.data_provider = ApiDataProvider(suites_data, environment.case_fields, environment.run_description)
+        self.data_provider = ApiDataProvider(suites_data, environment.case_fields, environment.run_description, environment.result_fields)
         self.suites_data_from_provider = self.data_provider.suites_input
         self.response_verifier = ApiResponseVerify(verify)
 
@@ -55,27 +56,26 @@ class ApiRequestHandler:
                 None
             )
             if automation_id_field:
-                context = automation_id_field["configs"][0]["context"]
-                if context["is_global"]:
-                    return None
-                elif project_id not in context["project_ids"]:
+                if automation_id_field["is_active"] is False:
                     return FAULT_MAPPING["automation_id_unavailable"]
+                for config in automation_id_field["configs"]:
+                    context = config["context"]
+                    if context["is_global"] or project_id in context["project_ids"]:
+                        return None
+                return FAULT_MAPPING["automation_id_unavailable"]
+            else:
+                return FAULT_MAPPING["automation_id_unavailable"]
         else:
             return response.error_message
 
-    def get_project_id(self, project_name: str, project_id: int = None) -> ProjectData:
+    def get_project_data(self, project_name: str, project_id: int = None) -> ProjectData:
         """
         Send get_projects with project name
         :project_name: Project name
         :returns: ProjectData
         """
-        response = self.client.send_get("get_projects")
-        if not response.error_message:
-            if isinstance(response.response_text, dict):  # Endpoints with pagination
-                projects_data = response.response_text["projects"]
-            else:  # Endpoints without pagination (legacy)
-                projects_data = response.response_text
-
+        projects_data, error = self.__get_all_projects()
+        if not error:
             available_projects = [
                 project
                 for project in projects_data
@@ -86,7 +86,7 @@ class ApiRequestHandler:
                 return ProjectData(
                     project_id=int(available_projects[0]["id"]),
                     suite_mode=int(available_projects[0]["suite_mode"]),
-                    error_message=response.error_message,
+                    error_message=error,
                 )
             elif len(available_projects) > 1:
                 if project_id in [project["id"] for project in available_projects]:
@@ -98,7 +98,7 @@ class ApiRequestHandler:
                     return ProjectData(
                         project_id=int(available_projects[project_index]["id"]),
                         suite_mode=int(available_projects[project_index]["suite_mode"]),
-                        error_message=response.error_message,
+                        error_message=error,
                     )
                 else:
                     return ProjectData(
@@ -116,7 +116,7 @@ class ApiRequestHandler:
             return ProjectData(
                 project_id=ProjectErrors.other_error,
                 suite_mode=-1,
-                error_message=response.error_message,
+                error_message=error,
             )
 
     def check_suite_id(self, project_id: int) -> (bool, str):
@@ -193,7 +193,7 @@ class ApiRequestHandler:
         add_suite_data = self.data_provider.add_suites_data()
         responses = []
         error_message = ""
-        for body in add_suite_data["bodies"]:
+        for body in add_suite_data:
             response = self.client.send_post(f"add_suite/{project_id}", body)
             if not response.error_message:
                 responses.append(response)
@@ -256,7 +256,7 @@ class ApiRequestHandler:
         add_sections_data = self.data_provider.add_sections_data()
         responses = []
         error_message = ""
-        for body in add_sections_data["bodies"]:
+        for body in add_sections_data:
             response = self.client.send_post(f"add_section/{project_id}", body)
             if not response.error_message:
                 responses.append(response)
@@ -288,16 +288,18 @@ class ApiRequestHandler:
         :project_id: project_id
         :returns: Tuple with list test case ID missing and error string.
         """
+        missing_cases_number = 0
         suite_id = self.suites_data_from_provider.suite_id
         returned_cases, error_message = self.__get_all_cases(project_id, suite_id)
-        if not error_message:
+        if error_message:
+            return False, error_message
+        if self.environment.case_matcher == MatchersParser.AUTO:
             test_cases_by_aut_id = {}
             for case in returned_cases:
                 aut_case_id = case["custom_automation_id"]
                 aut_case_id = aut_case_id if not aut_case_id else html.unescape(case["custom_automation_id"])
                 test_cases_by_aut_id[aut_case_id] = case
             test_case_data = []
-            missing_cases_number = 0
             for section in self.suites_data_from_provider.testsections:
                 for test_case in section.testcases:
                     if test_case.custom_automation_id in test_cases_by_aut_id.keys():
@@ -306,15 +308,29 @@ class ApiRequestHandler:
                             "case_id": case["id"],
                             "section_id": case["section_id"],
                             "title": case["title"],
+                            "custom_automation_id": test_case.custom_automation_id
                         })
                     else:
                         missing_cases_number += 1
             self.data_provider.update_data(case_data=test_case_data)
             if missing_cases_number:
-                self.environment.log(f"Found test cases not matching any TestRail case (count: {missing_cases_number})")
-            return missing_cases_number > 0, error_message
+                self.environment.log(f"Found {missing_cases_number} test cases not matching any TestRail case.")
         else:
-            return False, error_message
+            nonexistent_ids = []
+            all_case_ids = [case["id"] for case in returned_cases]
+            for section in self.suites_data_from_provider.testsections:
+                for test_case in section.testcases:
+                    if not test_case.case_id:
+                        missing_cases_number += 1
+                    elif int(test_case.case_id) not in all_case_ids:
+                        nonexistent_ids.append(test_case.case_id)
+            if missing_cases_number:
+                self.environment.log(f"Found {missing_cases_number} test cases without case ID in the report file.")
+            if nonexistent_ids:
+                self.environment.elog(f"Nonexistent case IDs found in the report file: {nonexistent_ids}")
+                return False, "Case IDs not in TestRail project or suite were detected in the report file."
+
+        return missing_cases_number > 0, ""
 
     def add_cases(self) -> (List[dict], str):
         """
@@ -326,16 +342,15 @@ class ApiRequestHandler:
         responses = []
         error_message = ""
         with self.environment.get_progress_bar(
-            results_amount=len(add_case_data["bodies"]), prefix="Adding test cases"
+            results_amount=len(add_case_data), prefix="Adding test cases"
         ) as progress_bar:
             with ThreadPoolExecutor(max_workers=MAX_WORKERS_ADD_CASE) as executor:
                 futures = {
                     executor.submit(
-                        self.client.send_post,
-                        f"add_case/{body.pop('section_id')}",
+                        self._add_case_and_update_data,
                         body,
                     ): body
-                    for body in add_case_data["bodies"]
+                    for body in add_case_data
                 }
                 responses, error_message = self.handle_futures(
                     futures=futures, action_string="add_case", progress_bar=progress_bar
@@ -348,43 +363,20 @@ class ApiRequestHandler:
             {
                 "case_id": response.response_text["id"],
                 "section_id": response.response_text["section_id"],
-                "title": response.response_text["title"],
+                "title": response.response_text["title"]
             }
             for response in responses
         ]
-        self.data_provider.update_data(case_data=returned_resources) if len(
-            returned_resources
-        ) > 0 else "Update skipped"
         return returned_resources, error_message
 
-    def update_case_result(self, run_id, case_id):
-        """
-        Updates result for case.
-        :run_id: run id
-        :case_id: case id
-        :returns: Tuple with dict (updated resource) and error string.
-        """
-        update_case_data = self.data_provider.add_result_for_case(case_id=case_id)
-        response_text = ""
-        if update_case_data:
-            response = self.client.send_post(
-                f"add_result_for_case/{run_id}/{update_case_data.pop('case_id')}",
-                update_case_data,
-            )
-            response_text = response.response_text
-            error_message = response.error_message
-        else:
-            error_message = FAULT_MAPPING["mismatch_between_case_id_and_result_file"]
-        return response_text, error_message
-
-    def add_run(self, project_id: int, run_name: str) -> (List[dict], str):
+    def add_run(self, project_id: int, run_name: str, milestone_id: int = None) -> (List[dict], str):
         """
         Creates a new test run.
         :project_id: project_id
         :run_name: run name
         :returns: Tuple with run id and error string.
         """
-        add_run_data = self.data_provider.add_run(run_name)
+        add_run_data = self.data_provider.add_run(run_name, milestone_id=milestone_id)
         response = self.client.send_post(f"add_run/{project_id}", add_run_data)
         return response.response_text.get("id"), response.error_message
 
@@ -450,7 +442,15 @@ class ApiRequestHandler:
             for test_result in results_data_chunk["results"]:
                 if test_result["attachments"]:
                     report_results_w_attachments.append(test_result)
-        self.upload_attachments(report_results_w_attachments, results, run_id)
+        if report_results_w_attachments:
+            attachments_count = 0
+            for result in report_results_w_attachments:
+                attachments_count += len(result["attachments"])
+            self.environment.log(f"Uploading {attachments_count} attachments "
+                                 f"for {len(report_results_w_attachments)} test results.")
+            self.upload_attachments(report_results_w_attachments, results, run_id)
+        else:
+            self.environment.log(f"No attachments found to upload.")
         return responses, error_message, progress_bar.n
 
     def handle_futures(self, futures, action_string, progress_bar):
@@ -465,6 +465,9 @@ class ApiRequestHandler:
                     if action_string == "add_results":
                         progress_bar.update(len(arguments["results"]))
                     else:
+                        if action_string == "add_case":
+                            arguments = arguments.to_dict()
+                            arguments.pop("case_id")
                         if not self.response_verifier.verify_returned_data(
                             arguments, response.response_text
                         ):
@@ -554,6 +557,17 @@ class ApiRequestHandler:
                     responses.append(response)
         return responses
 
+    def _add_case_and_update_data(self, case: TestRailCase):
+        case_body = case.to_dict()
+        if self.environment.case_matcher != MatchersParser.AUTO:
+            case_body.pop("custom_automation_id")
+        response = self.client.send_post(f"add_case/{case_body.pop('section_id')}", case_body)
+        if response.status_code == 200:
+            case.case_id = response.response_text["id"]
+            case.result.case_id = response.response_text["id"]
+            case.section_id = response.response_text["section_id"]
+        return response
+
     def __cancel_running_futures(self, futures, action_string):
         self.environment.log(
             f"\nAborting: {action_string}. Trying to cancel scheduled tasks."
@@ -565,7 +579,6 @@ class ApiRequestHandler:
         """
         Get all cases from all pages
         """
-        url = f"get_cases/{project_id}&suite_id={suite_id}"
         return self.__get_all_entities('cases', f"get_cases/{project_id}&suite_id={suite_id}")
 
     def __get_all_sections(self, project_id=None, suite_id=None) -> (List[dict], str):
@@ -580,6 +593,12 @@ class ApiRequestHandler:
         """
         return self.__get_all_entities('tests', f"get_tests/{run_id}")
 
+    def __get_all_projects(self) -> (List[dict], str):
+        """
+        Get all cases from all pages
+        """
+        return self.__get_all_entities('projects', f"get_projects")
+
     def __get_all_entities(self, entity: str, link=None, entities=[]) -> (List[dict], str):
         """
         Get all entities from all pages if number of entities is too big to return in single response.
